// Decompiled by Jad v1.5.8e. Copyright 2001 Pavel Kouznetsov.
// Jad home page: http://www.geocities.com/kpdus/jad.html
// Decompiler options: packimports(3) deadcode 
// Source File Name:   WaypointSystem.java

import java.io.*;
import java.util.*;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

public class WaypointSystem
    implements ISaveGame
{

    static WaypointSystem LoadWaypoints(NewAI ai)
    {
        String sDirectory = ai.GetModule().GetModulePath();
        File directory = new File(sDirectory);
        if(!directory.exists() || !directory.isDirectory())
        {
            GameObject.CrashLog("WaypointSystem.LoadWaypoints(): Directory " + sDirectory + "doesn't exist!!!\n");
            return null;
        }
        String fileName = ai.GetName() + "_" + ai.GetUniqueID() + ".wp";
        String filePath = sDirectory + fileName;
        String zipPath = sDirectory + "Waypoints.zip";
        WaypointSystem retWS = null;
        File wpFile = new File(filePath);
        try
        {
            ZipFile zipFile = new ZipFile(zipPath);
            Enumeration entries = zipFile.entries();
            ZipEntry foundEntry = null;
            while(entries.hasMoreElements()) 
            {
                ZipEntry entry = (ZipEntry)entries.nextElement();
                if(entry.getName().equalsIgnoreCase(fileName))
                {
                    foundEntry = entry;
                    break;
                }
            }
            if(foundEntry != null && wpFile.exists() && foundEntry.getTime() < wpFile.lastModified())
            {
                GameObject.CrashLog("Waypoints in zip file older than on disk\n");
                foundEntry = null;
            }
            if(foundEntry != null)
            {
                DataInputStream dataInput = new DataInputStream(zipFile.getInputStream(foundEntry));
                retWS = LoadWaypointsFromStream(ai, dataInput, fileName);
            }
        }
        catch(IOException e)
        {
            GameObject.CrashLog("ERROR reading waypoints (" + fileName + ") from " + "Waypoints.zip" + " file!\n");
        }
        try
        {
            if(retWS == null)
                if(wpFile.exists())
                {
                    DataInputStream dataInput = new DataInputStream(new FileInputStream(filePath));
                    retWS = LoadWaypointsFromStream(ai, dataInput, fileName);
                } else
                {
                    GameObject.CrashLog("Can't find file " + filePath + " on disk!\n");
                }
        }
        catch(IOException e)
        {
            GameObject.CrashLog("ERROR reading waypoints from file: " + filePath + "!\n");
        }
        return retWS;
    }

    void GetCloserPoint(Vector vPos)
    {
        short nearest = m_CurrWaypoint;
        AIWayPoint curr = (AIWayPoint)pts.get(m_CurrWaypoint);
        Vector vDist = new Vector(curr.m_vPos);
        vDist.Sub(vPos);
        vDist.fY = 0.0F;
        float fMinDist = vDist.Len2();
        for(int i = 0; i < curr.m_Neighbours.length; i++)
        {
            vDist.Set(((AIWayPoint)pts.get(curr.GetNeighbour(i))).m_vPos);
            vDist.Sub(vPos);
            vDist.fY = 0.0F;
            float fDist = vDist.Len2();
            if(fDist < fMinDist)
            {
                nearest = curr.GetNeighbour(i);
                fMinDist = fDist;
            }
        }

        SetCurrWaypoint(nearest);
    }

    void SetCurrWaypoint(short wp)
    {
        m_CurrWaypoint = wp;
    }

    static boolean InBuilding(ChromeActor actor)
    {
        actor.GetPosition(sv_vStart);
        sv_vEnd.Set(sv_vStart);
        sv_vStart.fY += 25F;
        sv_vEnd.fY -= 100F;
        Collision stCollInfo = actor.GetModule().TraceObjectsExclude(sv_vStart, sv_vEnd, actor);
        return stCollInfo != null && stCollInfo.Obstacle != null;
    }

    public void SGUpdate()
    {
        FileChunk.Log("[WaypointSystem.SGUpdate]\n");
    }

    WaypointSystem()
    {
        pts = new ArrayList();
        m_CurrWaypoint = 0;
        m_Owner = null;
        m_aPath = null;
        m_aWayPointsPath = null;
    }

    WaypointSystem(NewAI ai)
    {
        pts = new ArrayList();
        m_CurrWaypoint = 0;
        m_Owner = null;
        m_aPath = null;
        m_aWayPointsPath = null;
        m_Owner = ai;
        m_aWayPointsPath = new short[_MAX_MOVE_STEPS];
        m_aPath = new Vector[_MAX_MOVE_STEPS + 1];
        for(int i = 0; i < m_aPath.length; i++)
            m_aPath[i] = new Vector();

    }

    Vector[] GetPathTab()
    {
        return m_aPath;
    }

    static boolean IsOnMovingPlatform(ChromeActor actor)
    {
        actor.GetPosition(sv_vStart);
        sv_vEnd.Set(sv_vStart);
        sv_vStart.fY += 25F;
        sv_vEnd.fY -= 100F;
        Collision stCollInfo = actor.GetModule().TraceObjectsExclude(sv_vStart, sv_vEnd, actor);
        return stCollInfo != null && (stCollInfo.Obstacle instanceof LiftPlatform);
    }

    Vector GetCurrWPPos()
    {
        return ((AIWayPoint)pts.get(m_CurrWaypoint)).m_vPos;
    }

    private short InTable(Vector vPoint)
    {
        for(short j = 0; j < pts.size(); j++)
        {
            AIWayPoint pt = (AIWayPoint)pts.get(j);
            if(Math.abs(pt.m_vPos.fX - vPoint.fX) < AIWayPoint._IDENTITY_DIST_2 && Math.abs(pt.m_vPos.fY - vPoint.fY) < _CORRECTION && Math.abs(pt.m_vPos.fZ - vPoint.fZ) < AIWayPoint._IDENTITY_DIST_2)
                return j;
        }

        return -1;
    }

    public void GenerateWaypoints()
    {
        GenerateWaypoints(_MAX_TEST_VECTOR_LEN);
    }

    void GenerateWaypoints(float moveLen)
    {
        pts = null;
        pts = new ArrayList();
        Vector vS = new Vector();
        Vector vE = new Vector();
        Vector vectors[] = new Vector[8];
        vectors[0] = new Vector(0.0F, 0.0F, moveLen);
        vectors[1] = new Vector(moveLen, 0.0F, 0.0F);
        vectors[2] = new Vector(0.0F, 0.0F, -moveLen);
        vectors[3] = new Vector(-moveLen, 0.0F, 0.0F);
        vectors[4] = new Vector(moveLen, 0.0F, moveLen);
        vectors[5] = new Vector(-moveLen, 0.0F, moveLen);
        vectors[6] = new Vector(moveLen, 0.0F, -moveLen);
        vectors[7] = new Vector(-moveLen, 0.0F, -moveLen);
        short iCurrInd = 0;
        AIWayPoint currWP = null;
        m_Owner.SetCollAction(0);
        AIWayPoint startPoint = new AIWayPoint(m_Owner.GetPosition());
        pts.add(startPoint);
        SetCurrWaypoint((short)0);
        boolean bGenerate;
        for(bGenerate = true; iCurrInd < pts.size() && bGenerate; iCurrInd++)
        {
            currWP = (AIWayPoint)pts.get(iCurrInd);
            vS.Set(currWP.m_vPos);
            for(int i = 0; i < vectors.length; i++)
            {
                m_Owner.SetPositionWorld(vS);
                vE.Set(vS);
                Vector vMove = new Vector(vectors[i]);
                if(m_Owner.MoveByEllipsoidPure(vMove))
                {
                    vE.Add(vMove);
                    m_Owner.SetPositionWorld(vE);
                    if(InBuilding(m_Owner) && !IsOnMovingPlatform(m_Owner) && CorrectPosInBuilding(m_Owner, 0.0F))
                    {
                        vMove.Set(vS);
                        vMove.Sub(vE);
                        if(m_Owner.MoveByEllipsoidPure(vMove))
                        {
                            vE.Set(m_Owner.GetPosition());
                            short iwp = InTable(vE);
                            if(iwp == -1)
                            {
                                AIWayPoint wp = new AIWayPoint(vE);
                                iwp = (short)pts.size();
                                pts.add(wp);
                                if(pts.size() > _MAX_GEN_POINTS)
                                    bGenerate = false;
                            }
                            currWP.AddNeighbour(iwp);
                            ((AIWayPoint)pts.get(iwp)).AddNeighbour(iCurrInd);
                        }
                    }
                }
            }

        }

        if(!bGenerate)
            GameObject.CrashLog("max num of waypoints reached...");
        m_Owner.SetPositionWorld(startPoint.m_vPos);
        if(pts.size() < _MIN_GEN_POINTS_COUNT && moveLen > _MIN_TEST_VECTOR_LEN)
        {
            pts = null;
            GenerateWaypoints(_MIN_TEST_VECTOR_LEN);
            return;
        } else
        {
            SaveWaypoints(pts, m_Owner);
            m_Owner.SetCollAction(16);
            return;
        }
    }

    static WaypointSystem LoadWaypointsFromStream(NewAI ai, DataInputStream inputStream, String friendlyName)
    {
        WaypointSystem ws = new WaypointSystem(ai);
        try
        {
            Vector vActorPos = new Vector();
            vActorPos.ReadFromDisk(inputStream);
            if(!AIWayPoint.Equals(ai.GetPosition(), vActorPos))
                return null;
            int iPoints = inputStream.readInt();
            ws.pts = new ArrayList();
            for(int i = 0; i < iPoints; i++)
            {
                AIWayPoint pt = new AIWayPoint();
                pt.Load(inputStream);
                ws.pts.add(pt);
            }

            for(int i = 0; i < iPoints; i++)
            {
                AIWayPoint pt = (AIWayPoint)ws.pts.get(i);
                pt.LoadReferences(inputStream);
            }

            boolean bFound = false;
            for(int i = 0; i < iPoints; i++)
            {
                AIWayPoint pt = (AIWayPoint)ws.pts.get(i);
                if(pt.m_Neighbours == null || pt.m_Neighbours.length == 0)
                    bFound = true;
            }

            if(bFound)
                if(iPoints > 1)
                    GameObject.CrashLog("WARNING!!! Waypoint without exit in file " + friendlyName + ". Delete the file!!!\n");
                else
                if(iPoints == 1)
                    GameObject.CrashLog("WARNING!!! File " + friendlyName + " contains only one waypoint. AI won't be able to move!!!\n");
            ws.SetCurrWaypoint((short)0);
        }
        catch(Exception e)
        {
            GameObject.CrashLog("error in waypoints reading!!!\n");
            return null;
        }
        return ws;
    }

    static boolean CorrectPosInBuilding(ChromeActor actor, float fAdditionalUp)
    {
        Vector vMove = new Vector(0.0F, fAdditionalUp, 0.0F);
        vMove.Add(actor.GetPosition());
        actor.SetPositionWorld(vMove);
        vMove.Set(0.0F, -125F, 0.0F);
        if(actor.MoveByEllipsoidPure(vMove))
            return false;
        float fLen = vMove.Len();
        if(fLen == 0.0F)
        {
            vMove.Set(0.0F, _CORRECTION, 0.0F);
        } else
        {
            fLen -= _CORRECTION;
            vMove.Set(0.0F, -fLen, 0.0F);
        }
        vMove.Add(actor.GetPosition());
        actor.SetPositionWorld(vMove);
        return true;
    }

    static void VisualizeWaypointSystem(ChromeActor actor)
    {
        if(!(actor instanceof NewAI))
            return;
        NewAI ai = (NewAI)actor;
        if(ai.m_WpSystem == null)
            return;
        if(visibleWaypointsSystems == null)
            visibleWaypointsSystems = new ArrayList();
        VisibleWaypointsHolder holder = new VisibleWaypointsHolder(ai);
        for(int i = 0; i < ai.m_WpSystem.pts.size(); i++)
        {
            MeshObject cMesh = (MeshObject)actor.CreateObject("MeshObject");
            cMesh.LoadMesh("Data/Interface/zarowka.3da");
            cMesh.ActivateOnModule();
            cMesh.ClearFlag(1);
            cMesh.SetLighting(false);
            cMesh.EnableRendering(true);
            cMesh.SetFlag(2048);
            cMesh.SetPositionWorld(((AIWayPoint)ai.m_WpSystem.pts.get(i)).m_vPos);
            cMesh.SetName("Zarowka" + i);
            holder.AddMesh(cMesh);
        }

        visibleWaypointsSystems.add(holder);
        aRoute = new MeshObject[_MAX_MOVE_STEPS + 1];
        for(int i = 0; i < aRoute.length; i++)
        {
            MeshObject cMesh = (MeshObject)actor.CreateObject("MeshObject");
            cMesh.LoadMesh("data/interface/choragiewka.3da");
            cMesh.ActivateOnModule();
            cMesh.ClearFlag(1);
            cMesh.SetFlag(2048);
            cMesh.SetLighting(false);
            cMesh.EnableRendering(false);
            cMesh.SetScale(0.3F, 0.3F, 0.3F);
            aRoute[i] = cMesh;
        }

    }

    FindPathInfo FindPath(Vector vStart, Vector vEnd)
    {
        int iSteps = 0;
        AIWayPoint currWP = (AIWayPoint)pts.get(m_CurrWaypoint);
        AIWayPoint currNeighbour = null;
        short chosenNeighbour = -1;
        m_aPath[0].Set(currWP.m_vPos);
        m_aWayPointsPath[0] = m_CurrWaypoint;
        m_CalcSet.clear();
        m_CalcSet.add(currWP);
        iSteps++;
        if(aRoute != null)
        {
            for(int i = 0; i < aRoute.length; i++)
                aRoute[i].EnableRendering(false);

        }
        for(; iSteps < m_aWayPointsPath.length; iSteps++)
        {
            float fMinDist = MathUtils.dist2D2(currWP.m_vPos, vEnd);
            chosenNeighbour = -1;
            if(currWP.m_Neighbours == null)
            {
                m_FindPathInfo.iWPSteps = iSteps;
                m_FindPathInfo.iRealSteps = iSteps;
                return m_FindPathInfo;
            }
            for(int i = 0; i < currWP.m_Neighbours.length; i++)
            {
                currNeighbour = (AIWayPoint)m_Owner.m_WpSystem.pts.get(currWP.m_Neighbours[i]);
                float fDist = MathUtils.dist2D2(currNeighbour.m_vPos, vEnd);
                if(fDist < fMinDist && !m_CalcSet.contains(currNeighbour))
                {
                    fMinDist = fDist;
                    chosenNeighbour = currWP.m_Neighbours[i];
                }
            }

            if(chosenNeighbour == -1)
                break;
            currWP = (AIWayPoint)pts.get(chosenNeighbour);
            m_CalcSet.add(currWP);
            m_aPath[iSteps].Set(currWP.m_vPos);
            m_aWayPointsPath[iSteps] = chosenNeighbour;
            if(aRoute != null)
            {
                aRoute[iSteps].EnableRendering(true);
                aRoute[iSteps].SetPositionWorld(currWP.m_vPos);
            }
        }

        m_FindPathInfo.iWPSteps = iSteps;
        if(CanMakeExactMove(currWP, vEnd))
        {
            m_aPath[iSteps].Set(vEnd);
            if(aRoute != null)
            {
                aRoute[iSteps].EnableRendering(true);
                Vector vPos = new Vector(vEnd);
                vPos.fY = m_aPath[iSteps - 1].fY;
                aRoute[iSteps].SetPositionWorld(vPos);
            }
            iSteps++;
        }
        m_FindPathInfo.iRealSteps = iSteps;
        return m_FindPathInfo;
    }

    void SaveWaypoints(ArrayList pts, ChromeActor actor)
    {
        String sDirectory = actor.GetModule().GetModulePath();
        File directory = new File(sDirectory);
        if(!directory.exists() || !directory.isDirectory())
            directory.mkdir();
        String fileName = sDirectory + actor.GetName() + "_" + actor.GetUniqueID() + ".wp";
        GameObject.CrashLog("Saving waypoints:" + fileName + "...");
        try
        {
            DataOutputStream output = new DataOutputStream(new FileOutputStream(fileName));
            actor.GetPosition().WriteToDisk(output);
            output.writeInt(pts.size());
            for(int i = 0; i < pts.size(); i++)
            {
                AIWayPoint pt = (AIWayPoint)pts.get(i);
                pt.Save(output);
            }

            for(int i = 0; i < pts.size(); i++)
            {
                AIWayPoint pt = (AIWayPoint)pts.get(i);
                pt.SaveReferences(output);
            }

            output.close();
            GameObject.CrashLog("Done.\n");
        }
        catch(Exception e)
        {
            GameObject.CrashLog("error in waypoints writing!!!\n");
            e.printStackTrace();
            return;
        }
    }

    short FindNearestPoint(Vector vPos)
    {
        float minDist2 = 3.402823E+38F;
        short pointIdx = -1;
        short ptsNum = (short)pts.size();
        for(short i = 0; i < ptsNum; i++)
        {
            AIWayPoint point = (AIWayPoint)pts.get(i);
            float dist2;
            if(point != null && (dist2 = MathUtils.dist2D2(vPos, point.m_vPos)) < minDist2)
            {
                minDist2 = dist2;
                pointIdx = i;
            }
        }

        return pointIdx;
    }

    static void DevisualizeWaypointSystem(ChromeActor actor)
    {
        if(!(actor instanceof NewAI))
            return;
        if(visibleWaypointsSystems == null)
            return;
        NewAI ai = (NewAI)actor;
        VisibleWaypointsHolder holder = null;
        MeshObject mesh = null;
        for(int i = 0; i < visibleWaypointsSystems.size(); i++)
        {
            holder = (VisibleWaypointsHolder)visibleWaypointsSystems.get(i);
            if(holder.GetAI() == ai)
            {
                for(int j = 0; j < holder.GetMeshesNum(); j++)
                {
                    mesh = holder.GetMesh(j);
                    mesh.delete();
                    mesh = null;
                }

                holder.delete();
                holder = null;
                visibleWaypointsSystems.remove(i);
                return;
            }
        }

    }

    public void SGSaveChunk(FileChunk cFCParent)
    {
        FileChunk.Log("[WaypointSystem.SGSaveChunk]\n");
        FileChunk cFC = GameObject.NewChunk(1609, cFCParent);
        cFC.SaveInt(m_CurrWaypoint);
        cFC.SaveObjectRef(m_Owner);
        if(m_aPath == null)
        {
            cFC.SaveInt(-1);
        } else
        {
            cFC.SaveInt(m_aPath.length);
            for(int i = 0; i < m_aPath.length; i++)
                cFC.SaveVector(m_aPath[i]);

        }
        if(m_aWayPointsPath == null)
        {
            cFC.SaveInt(-1);
        } else
        {
            cFC.SaveInt(m_aWayPointsPath.length);
            for(int i = 0; i < m_aWayPointsPath.length; i++)
                cFC.SaveInt(m_aWayPointsPath[i]);

        }
        cFC.SaveInt(m_FindPathInfo.iRealSteps);
        cFC.SaveInt(m_FindPathInfo.iWPSteps);
        cFC.delete();
    }

    public void SGLoadChunk(FileChunk cParentFC)
    {
        FileChunk.Log("[WaypointSystem.SGLoadChunk]\n");
        FileChunk cFC = GameObject.LoadChunk(cParentFC);
        if(cFC.GetID() == 1609)
        {
            m_CurrWaypoint = (short)cFC.LoadInt();
            m_Owner = (NewAI)cFC.LoadObjectRef();
            int iTemp = cFC.LoadInt();
            if(iTemp < 0)
            {
                m_aPath = null;
            } else
            {
                m_aPath = new Vector[iTemp];
                for(int i = 0; i < iTemp; i++)
                    m_aPath[i] = cFC.LoadVector();

            }
            iTemp = cFC.LoadInt();
            if(iTemp < 0)
            {
                m_aWayPointsPath = null;
            } else
            {
                m_aWayPointsPath = new short[iTemp];
                for(int i = 0; i < iTemp; i++)
                    m_aWayPointsPath[i] = (short)cFC.LoadInt();

            }
            m_FindPathInfo.iRealSteps = cFC.LoadInt();
            m_FindPathInfo.iWPSteps = cFC.LoadInt();
        }
        cFC.delete();
    }

    boolean CanMakeExactMove(AIWayPoint wp, Vector vEnd)
    {
        if(MathUtils.dist2D2(wp.m_vPos, vEnd) < 1.0F)
            return false;
        Vector vDir = new Vector(vEnd);
        vDir.Sub(wp.m_vPos);
        vDir.fY = 0.0F;
        vDir.Normalize();
        Vector vNeighbourDir = new Vector();
        iNdx = 0;
        for(int i = 0; i < wp.m_Neighbours.length; i++)
        {
            vNeighbourDir.Set(((AIWayPoint)pts.get(wp.m_Neighbours[i])).m_vPos);
            vNeighbourDir.Sub(wp.m_vPos);
            vNeighbourDir.fY = 0.0F;
            vNeighbourDir.Normalize();
            if(vDir.Dot(vNeighbourDir) > 0.7072F)
            {
                if(iNdx == 2)
                {
                    GameObject.CrashLog("\n\n\nBLAD W CANMAKEEXACTMOVE!!!\n");
                    GameObject.CrashLog("wp.m_vPos:" + wp.m_vPos + "\n");
                    GameObject.CrashLog("vEnd:" + vEnd + "\n");
                    for(int k = 0; k < wp.m_Neighbours.length; k++)
                        GameObject.CrashLog("NEIGHBOUR[" + k + "]:" + ((AIWayPoint)pts.get(wp.m_Neighbours[k])).m_vPos + "\n");

                    GameObject.CrashLog("\n\n\n");
                    return false;
                }
                vTemp[iNdx] = ((AIWayPoint)pts.get(wp.m_Neighbours[i])).m_vPos;
                iNdx++;
            }
        }

        if(iNdx < 2)
            return false;
        vDir.Set(vTemp[0]);
        vDir.Sub(vTemp[1]);
        vDir.fY = 0.0F;
        vDir.Normalize();
        vNeighbourDir.Set(vEnd);
        vNeighbourDir.Sub(vTemp[1]);
        vNeighbourDir.fY = 0.0F;
        vNeighbourDir.Normalize();
        float fCross = vDir.fZ * vNeighbourDir.fX - vNeighbourDir.fZ * vDir.fX;
        vNeighbourDir.Set(wp.m_vPos);
        vNeighbourDir.Sub(vTemp[1]);
        vNeighbourDir.fY = 0.0F;
        vNeighbourDir.Normalize();
        return (vDir.fZ * vNeighbourDir.fX - vNeighbourDir.fZ * vDir.fX) * fCross >= 0.0F;
    }

    static final String _ZIP_FILE_NAME = "Waypoints.zip";
    static float _CORRECTION = 5F;
    static float _MAX_TEST_VECTOR_LEN = 150F;
    static float _MIN_TEST_VECTOR_LEN = 50F;
    static int _MIN_GEN_POINTS_COUNT = 50;
    static int _MAX_GEN_POINTS = 4000;
    ArrayList pts;
    short m_CurrWaypoint;
    NewAI m_Owner;
    static int _MAX_MOVE_STEPS;
    static HashSet m_CalcSet;
    Vector m_aPath[];
    short m_aWayPointsPath[];
    static FindPathInfo m_FindPathInfo = new FindPathInfo();
    static Vector vTemp[] = new Vector[2];
    static int iNdx = 0;
    private static Vector sv_vStart = new Vector();
    private static Vector sv_vEnd = new Vector();
    static ArrayList visibleWaypointsSystems = null;
    static MeshObject aRoute[] = null;

    static 
    {
        _MAX_MOVE_STEPS = 10;
        m_CalcSet = new HashSet(_MAX_MOVE_STEPS);
    }
}
